import json
import os
from typing import Any, Iterator
from urllib.parse import urljoin

import backoff
import requests

# https://developers.notion.com/reference/intro#conventions
NOTION_BASE_URL = 'https://api.notion.com'

# https://developers.notion.com/reference/versioning
NOTION_VERSION = '2022-02-22'


def make_params(vars):
    """quick hacky way to build body params
    vars is locals() passed from each api endpoint, eg query_database().
    This function copies every var if
        - it's not 'self' or 'kw'
        - not start with '_':
            in case any api endpoint code need local var, it should start with _
        - not end with '_id':
            all url params are also defined in api endpoint functions
            they all like '*_id', eg page_id, block_id etc
    The goal is to minimize api endpoint code, easier to read and match to official document.
    This is less hacky than using inspect :D and still kind of readable/maintainable
        and api endpoint code is minimal enough.
    """
    params = {
        name: vars[name]
        for name in vars
        if name not in ('self', 'kw')  # skip self and kw
        and name[0] != '_'   # skip possible local use vars
        and not name.endswith('_id')  # skip url params
        if vars.get(name) is not None
    }
    if 'kw' in vars:
        # copy everything under 'kw'
        params.update(vars['kw'])
    return params


class Client:
    """A simple direct translation of offical API reference https://developers.notion.com/reference/intro into api endpoints,
    NotionParams should have a helper for each api to fill-in options, so they can be used as
    ```
    client.append_block_children(block_id, **NP.append_markdown('markdown text'))
    ```
    """

    def __init__(self, token: str = None) -> None:
        if token is None:
            token = os.environ.get('NOTION_TOKEN')
        self._session = requests.Session()
        self._session.headers.update({
            "Accept": "application/json",
            "Notion-Version": NOTION_VERSION,
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}"
        })
        self._last_exc = None

    @backoff.on_exception(
        # retry 429 until succ or other error
        backoff.expo,
        requests.exceptions.HTTPError,
        giveup=lambda exc: exc.response is not None and exc.response.status_code != 429,
        max_value=10,
        jitter=None,
    )
    @backoff.on_exception(
        # retry 502 3 times
        backoff.constant,
        requests.exceptions.HTTPError,
        giveup=lambda exc: exc.response is not None and exc.response.status_code != 502,
        max_tries=3,
        interval=2,  # 2 seconds
        jitter=None,
    )
    def _request_core(self, url, **kw):
        r = self._session.request(url=urljoin(NOTION_BASE_URL, url), **kw)
        r.raise_for_status()
        return r.json()

    def _request(self, url, **kw):
        """save exception for diagnostic"""
        try:
            return self._request_core(url, **kw)
        except requests.exceptions.HTTPError as exc:
            self._last_exc = exc
            if str(os.environ.get('NOTION_PARAMS_SHOW_LAST_EXC') or '').lower() in ('yes', 'y', 'true', 't', '1'):
                self._show_last_exc()
            raise
    
    def _show_last_exc(self):
        exc = self._last_exc
        if not exc:
            return
        if exc.response is not None:
            print('Response', exc.response.status_code, exc.response.text)
        if exc.request is not None:
            print('Request body', exc.request.body)

    def _api(self, method, url, vars=None):
        params = make_params(vars) if vars else None
        return self._request(url=url, method=method, json=params)

    def _paginate(self, method, url, vars) -> Iterator[Any]:
        # https://developers.notion.com/reference/pagination
        """kw['json'] can have start_cursor, page_size, see sample query_database()"""
        params = make_params(vars)
        while True:
            response = self._request(url=url, method=method, json=params)
            yield from response.get('results') or []
            next_cursor = response.get('next_cursor')
            if not next_cursor:
                break
            params['start_cursor'] = next_cursor

    # the following 'api' endpoints are simple mirror of https://developers.notion.com/reference/intro
    # in the same order and naming conventions for easier jump to official documents

    def query_database(self, database_id, *, filter=None, sorts=None, start_cursor=None, page_size=None) -> Iterator[Any]:
        """https://developers.notion.com/reference/post-database-query"""
        yield from self._paginate('post', f'/v1/databases/{database_id}/query', locals())

    def create_database(self, *, parent, properties, title=None, icon=None, cover=None) -> dict:
        """https://developers.notion.com/reference/create-a-database"""
        # Note: 'icon' and 'cover' are not listed in document but showed in sample
        return self._api('post', '/v1/databases', locals())

    def update_database(self, database_id, *, title=None, properties=None):
        """https://developers.notion.com/reference/update-a-database"""
        return self._api('patch', f'/v1/databases/{database_id}', locals())

    def retrieve_database(self, database_id, ):
        """https://developers.notion.com/reference/retrieve-a-database"""
        return self._api('get', f'/v1/databases/{database_id}')

    def retrieve_page(self, page_id):
        """https://developers.notion.com/reference/retrieve-a-page"""
        return self._api('get', f'/v1/pages/{page_id}')

    def create_page(self, *, parent, properties, children=None, icon=None, cover=None):
        """https://developers.notion.com/reference/post-page"""
        return self._api('post', '/v1/pages', locals())

    def update_page(self, page_id, *, properties=None, archived=None, icon=None, cover=None):
        """https://developers.notion.com/reference/patch-page"""
        return self._api('patch', f'/v1/pages/{page_id}', locals())

    def retrieve_page_property_item(self, page_id, property_id, *, start_cursor=None, page_size=None):
        """https://developers.notion.com/reference/retrieve-a-page-property"""
        yield from self._paginate('get', f'/v1/pages/{page_id}/properties/{property_id}', locals())

    def retrieve_block(self, block_id):
        """https://developers.notion.com/reference/retrieve-a-block"""
        return self._api('get', f'/v1/blocks/{block_id}')

    def update_block(self, block_id, *, archived=None, **kw):
        """https://developers.notion.com/reference/update-a-block
        (kw is) the block object type value with the properties to be updated. Currently only text (for supported block types) and checked (for to_do blocks) fields can be updated.
        """
        return self._api('patch', f'/v1/blocks/{block_id}', {**locals(), **kw})

    def retrieve_block_children(self, block_id, *, start_cursor=None, page_size=None):
        """https://developers.notion.com/reference/get-block-children"""
        yield from self._paginate('get', f'/v1/blocks/{block_id}/children', locals())

    def append_block_children(self, block_id, *, children):
        """https://developers.notion.com/reference/patch-block-children"""
        return self._api('patch', f'/v1/blocks/{block_id}/children', locals())

    def delete_block(self, block_id):
        """https://developers.notion.com/reference/delete-a-block"""
        return self._api('delete', f'/v1/blocks/{block_id}')

    def retrieve_user(self, user_id):
        """https://developers.notion.com/reference/get-user"""
        return self._api('get', f'/v1/users/{user_id}')

    def list_users(self, *, start_cursor=None, page_size=None):
        """https://developers.notion.com/reference/get-users"""
        yield from self._paginate('get', f'/v1/users', locals())

    def retrieve_bot_user(self):
        """https://developers.notion.com/reference/get-self"""
        return self._api('get', '/v1/users/me')

    def search(self, *, query=None, sort=None, filter=None, start_cursor=None, page_size=None):
        """https://developers.notion.com/reference/post-search"""
        yield from self._paginate('post', '/v1/search', locals())



from typing import Any, List, Mapping

from .markdown import md, md_line
from .client import Client


class NotionParams:
    """
    Notion API reference https://developers.notion.com/reference/intro
    Find emoji https://emojipedia.org/people/

    ```
    from notion_params import NotionParams as NP
    notion = NP.get_client(token)  # or set token in env NOTION_TOKEN
    page_id = '7458781ba20644e0b85045209554ff3d'
    page = notion.retrieve_page(page_id=page_id)

    # create sub page, with markdown text
    sub_page = notion.create_page(**NP.create_page(
        page['id'], title='...', emoji='😆',
        text="markdown content", 
    )
    ```
    """

    md = md
    md_line = md_line

    @staticmethod
    def get_client(token=None):
        return Client(token=token)

    @staticmethod
    def create_page(parent_page_id: str, *, title: str, text: str = None, emoji: str = None):
        """
        ```
        sub_page = notion.pages.create(
            **NotionParams.create_page(
                page['id'],
                title='title string',
                emoji='👷',
                text="markdown...this will be the content",
            )
        )
        ```
        """
        params = {
            # parent https://developers.notion.com/reference/page#page-parent
            'parent': {
                'type': 'page_id',
                'page_id': parent_page_id,
            },
            'properties': {
                # properties https://developers.notion.com/reference/page#all-pages
                # when parent.type is "page_id", only valid is title
                # title can have markdown but displayed as normal text
                # here use md_line only to format notion text param
                'title': md_line(title),
            },
        }
        if text:
            data = md(text)
            # print('markdown', json.dumps(data, indent=4))
            params['children'] = data
        if emoji:
            params['icon'] = {'emoji': emoji}
        return params

    @staticmethod
    def update_page(*, title: str, emoji: str = None, archived=None):
        """
        ```
        notion.pages.update(
            sub_page['id'],
            **NotionParams.update_page(
                title='new title text',
                emoji='😃',
            )
        )
        ```
        """
        params = {
            'properties': {
                'title': md_line(title),
            },
        }
        if emoji:
            params['icon'] = {'emoji': emoji}
        if archived in (True, False):
            params['archived'] = archived
        return params

    @staticmethod
    def append_markdown(text: str):
        """options in https://developers.notion.com/reference/patch-block-children
        Usage: `client.append_block_children(block_id, **NP.append_markdown('markdown text'))`
        """
        return {
            'children': md(text),
        }

    @staticmethod
    def table_df(df, include_rows=0, has_row_header=True, has_column_header=True):
        """convert DataFrame to one table block
        :param include_rows: number of rows to include, default only include header
            too many rows may cause result too large
        ```
        result = notion.blocks.children.append(
            block_id=sub_page['id'],
            children=[NotionParams.table_df(df)],
        )
        # table = next((i for i in result['results'] if i['type'] == 'table'))  # assume only one table is created
        table = result['results'][0]  # or simpler because we know only one block is created
        for i in range(0, len(df), 100):
            rows = notion.blocks.children.append(
                block_id=table['id'],
                children=df[i:i+100]
            )
        ```
        """
        # https://developers.notion.com/reference/block#table-blocks
        header_row = [
            {
                "type": "table_row",
                "table_row": {
                    "cells": [
                        [{
                            "text": {"content": col}
                        }]
                        for col in df.columns
                    ]
                }
            }
        ] if has_column_header else []
        return {
            'type': 'table',
            'table': {
                "table_width": len(df.columns),
                "has_column_header": has_column_header,
                "has_row_header": has_row_header,
                "children": header_row + NotionParams.table_df_rows(df[:include_rows])
            }
        }

    @staticmethod
    def table_df_rows(df):
        """convert all rows to list of blocks, pass in slice if DataFrame too large
        sample
        ```
        notion.blocks.children.append(
            block_id=table_block_id,
            children=NotionParams.table_df_rows(df[100:200]),
        )
        ```
        """
        return df.apply(
            lambda row: {
                "type": "table_row",
                "table_row": {
                    "cells": [
                        [{
                            "text": {"content": str(row[col])}
                        }]
                        for col in df.columns
                    ]
                }
            } if len(row) else {},
            # NOTE: if df length is 0, lambda is invoked once with empty row
            # in this case if reference row[col] will cause confusing error
            # 'DataFrame' object has no attribute 'tolist'
            axis=1,
        ).tolist()

    @staticmethod
    def df_columns_add_prefix_for_database(df):
        width = (len(df.columns)-1)//10 + 1  # 2 cols -> 1 digits, 10 cols -> 1 digits, 11 cols -> 2 digits
        return df.rename(columns={
            col: f'{str(i).zfill(width)}-{col}'
            for i, col in enumerate(df.columns[1:])
        })

    @staticmethod
    def create_database(page_id, *, title: str, columns: List[str], column_types: Mapping[str, str] = None, emoji: str = None):
        # https://developers.notion.com/reference/create-a-database
        """
        :param column_types: custom column type for each column, if not defined, default is 'rich_text'

        demo:
        ```
        df = pd.DataFrame(...)
        db = notion.databases.create(
            **NotionParams.create_database(
                page['id'],
                title=db_name,
                columns=df.columns,
                emoji='😀',
            ),
        )
        print('created db', json.dumps(db, indent=4))
        for _idx, row in db.iterrows():
            notion.pages.create(**NotionParams.create_database_row(db['id'], row=row))
        # or
        # df.apply(
        #     lambda row: notion.pages.create(
        #         **NotionParams.create_database_row(
        #             db['id'],
        #             row=row,
        #             # emoji='😀',
        #         ),
        #     ),
        #     axis=1,  ### REMEMBER THIS ! to loop lines
        # )
        ```
        """
        column_types = {
            column: (column_types and column_types.get(
                column) or 'rich_text') if idx else 'title'
            for idx, column in enumerate(columns)
        }
        params = {
            'parent': {
                'type': 'page_id',
                'page_id': page_id,
            },
            'title': md_line(title),
            'properties': {
                column: {column_types[column]: {}}
                for column in columns
            },
        }
        if emoji:
            params['icon'] = {'emoji': emoji}
        # print(json.dumps(params, indent=4))
        return params

    @staticmethod
    def create_database_row(db_id: str, *, row: Mapping[str, Any], columns: List[str] = None, column_types: Mapping[str, str] = None, emoji: str = None):
        """
        ```
        for _idx, row in db.iterrows():
            notion.pages.create(**NotionParams.create_database_row(db['id'], row=row))
        ```
        """
        if columns is None:
            columns = row.keys()
        column_types = {
            column: (column_types and column_types.get(
                column) or 'rich_text') if idx else 'title'
            for idx, column in enumerate(columns)
        }
        params = {
            'parent': {
                'type': 'database_id',
                'database_id': db_id
            },
            'properties': {
                column: {
                    'type': column_types[column],
                    column_types[column]: row[column] if column_types[column] not in ('title', 'rich_text') else [{
                        # special value format for title, must be array
                        'text': {'content': str(row[column])},
                    }],
                }
                for column in columns
            },
            # every database row is a page, here it can have children list of blocks
            # 'children': [{block},...]
        }
        if emoji:
            params['icon'] = {'emoji': emoji}
        # print('create_database_row', json.dumps(params, indent=4))
        return params

    @staticmethod
    def find_child(response, *, type_: str, title: str = None):
        """find child in list children response or block append response"""
        if isinstance(response, list):
            # user can pass results array directly, ie NotionParams.find_child(response['results'], ...)
            children = response
        elif isinstance(response, dict):
            # should be normal response
            children = response.get('results') or []
        else:
            raise ValueError('unexpected response type')
        return next((
            i
            for i in children
            if i['type'] == type_
            if title is None or title == i[type_].get('title')
        ), None)



import json
import warnings

import marko
import pydash
from marko.ext.gfm.elements import Strikethrough

# see https://developers.notion.com/reference/rich-text#annotations
NOTION_COLOR_NAMES = "gray", "brown", "orange", "yellow", "green", "blue", "purple", "pink", "red"

def decode_style(code):
    """simple <span style> decoder
    :param code: inline html code, eg "<span style='color:blue'>" or "</span>" or any other inline html code
    :returns: 
        if style match, return notion color code, "red", "blue" etc, see NOTION_COLOR_NAMES
        if /span return '/'
        otherwise return ''
    """
    if not code:
        return ''
    code = code.lower().strip('<>')
    if code == '/span':
        return '/'
    if not code.startswith('span '):
        return ''
    # faster coding than regex :D
    while ' =' in code:
        code = code.replace(' =', '=')
    while '= ' in code:
        code = code.replace('= ', '=')
    style = next((i for i in code.split() if i.startswith('style=')), None)
    if not style:
        return ''
    style = style.split('=', 1)[-1].strip('"\'').split(':', 1)  # (color, blue)
    # possible values https://developers.notion.com/reference/rich-text#annotations
    if style[0] == 'color':
        color, is_background = style[1], False
    elif style[0] == 'background-color':
        color, is_background = style[1], True
    else:
        return ''
    if color not in NOTION_COLOR_NAMES:
        color = "blue"  # show color is working but not the desired one
    return color if not is_background else f'{color}_background'


class MarkoNotionRenderer(marko.Renderer):
    # mostly followed https://github.com/frostming/marko/blob/master/marko/md_renderer.py
    # there are a few known issues, see cmd_markdown_demo
    # so far, simple format works, avoid nesting blocks, ie, sub-list, sub-quote etc

    def render_children(self, element):
        # copy from Renderer.render_childern
        if isinstance(element, str):
            return [{'text': {'content': element}}]
        if isinstance(element.children, str):
            return [{'text': {'content': element.children}}]
        return [self.render(child) for child in element.children]

    def _render_as(self, type_, element, **kw):
        result = {
            "type": type_,
            type_: {} if element is None else {'rich_text': self.render_children(element), **kw}
        }

        # fix custom block type
        if type_ == 'paragraph' and result[type_].get('rich_text'):
            # scan for !!callout etc for custom paragraph
            text_list = result[type_].get('rich_text')
            line0 = pydash.get(text_list, '0.text.content')
            line1 = pydash.get(text_list, '1.text.content')
            if line0 and line0.startswith('!!') and line1 == '\n':
                custom_type, *args = line0[2:].split()
                args = {
                    k.strip(): v.strip()
                    for i in args
                    for k, v in [i.split('=', 1)]
                }
                # print('='*80, 'CUSTOM??', custom_type, args)
                # print(repr(line0))
                # print(repr(line1))
                if custom_type in ('callout', 'todo', 'to_do', 'toggle'):
                    if custom_type == 'todo':
                        custom_type = 'to_do'
                    result['type'] = custom_type
                    result[custom_type] = result.pop(type_)
                    result[custom_type]['rich_text'] = text_list[2:]
                    if args.get('emoji'):
                        pydash.set_(result[custom_type], 'icon.emoji', args['emoji'])
                    if args.get('color'):
                        pydash.set_(result[custom_type], 'color', args['color'])
                    type_ = custom_type

        # move 2nd level nested paragraph to rich_text array
        # move 2nd level other block type to children
        if result[type_] and type_ != 'paragraph':
            text_list = result[type_].get('rich_text')
            if text_list:
                children = [
                    i
                    for i in text_list
                    if i.get('type') and i.get('type') != 'paragraph'
                ]
                def copy_inner_paragraph():
                    for item in text_list:
                        if 'type' not in item:
                            yield item
                        elif item.get('type') == 'paragraph':
                            yield from item['paragraph']['rich_text']
                result[type_]['rich_text'] = list(copy_inner_paragraph())
                if children:
                    if type_ in ('quote',):
                        # types that allow children
                        result[type_]['children'] = children
                    else:
                        warnings.warn(f'Lost of inner blocks {[i["type"] for i in children]}')

        # fix color
        if type_ == 'paragraph' and result[type_].get('rich_text'):
            # scan for <span style='color|background-color:<color>'> and </span>
            text_list = result[type_].get('rich_text')
            # {"text": {"content": "<span style='color:blue'>"}, "annotations": {"code": true}},
            # {"text": {"content": "blue text"}},  <<< can be many in between
            # {"text": {"content": "</span>"}, "annotations": {"code": true}},  <<< don't consider nesting style
            def fix_color(text_list):
                colors = []
                for item in text_list:
                    if pydash.get(item, 'annotations.code'):
                        style = decode_style(pydash.get(item, 'text.content'))
                        if style == '/':
                            if colors:
                                colors.pop(-1)  # remove last
                                continue
                        elif style:
                            colors.append(style)  # support nesting color
                            continue
                        # otherwise pass through to be rendered as inline html
                    if colors:
                        pydash.set_(item, 'annotations.color', colors[-1])
                    yield item
            result[type_]['rich_text'] = list(fix_color(text_list))

        return result

    def render_paragraph(self, element):
        # https://developers.notion.com/reference/block#paragraph-blocks
        return self._render_as("paragraph", element)

    def _text(self, element, bold=False, italic=False, strikethrough=False, underline=False, code=False, color: str = 'default', url=None):
        """
        :param code: True for inline code
        :param color: see full list https://developers.notion.com/reference/rich-text#annotations
        """
        # https://developers.notion.com/reference/rich-text#annotations
        annotations = {}
        if bold:
            annotations['bold'] = True
        if italic:
            annotations['italic'] = True
        if strikethrough:
            annotations['strikethrough'] = True
        if underline:
            annotations['underline'] = True
        if code:
            annotations['code'] = True
        if color and color != 'default':
            annotations['color'] = color
        text = self.render_children(element)
        if isinstance(text, list):
            # test.content should be string, flatten nesting children here
            # don't consider complex nesting for now
            # print('text <- ', json.dumps(text, indent=4))
            text = ' '.join([i['text']['content'] for i in text])
            # print('text -> ', json.dumps(text, indent=4))
        result = {
            'text': {
                'content': text,
            }
        }
        if url:
            # https://developers.notion.com/reference/rich-text#link-objects
            pydash.set_(result, 'text.link.url', url)
        if annotations:
            result['annotations'] = annotations
        return result

    def render_strikethrough(self, element):
        return self._text(element, strikethrough=True)

    def render_list(self, element):
        # https://developers.notion.com/reference/block#numbered-list-item-blocks
        # https://developers.notion.com/reference/block#bulleted-list-item-blocks
        result = self.render_children(element)
        assert isinstance(result, list)
        type_ = 'numbered_list_item' if element.ordered else 'bulleted_list_item'
        # print('render_list', type_, json.dumps(result, indent=4))
        for item in result:
            t, item['type'] = item['type'], type_
            item[type_] = item.pop(t)
        # print('render_list patched to', json.dumps(result, indent=4))
        return result
        # return self._render_as(type_, element)

    def render_list_item(self, element):
        result = self.render_children(element)
        if not isinstance(result, list) and len(result) == 1:
            warnings.warn(f'unexpect list item longer than 1 {json.dumps(result, indent=4)}')
            raise NotImplementedError
        return result[0]

    def render_quote(self, element):
        # https://developers.notion.com/reference/block#quote-blocks
        return self._render_as('quote', element)

    def render_fenced_code(self, element):
        # code block that starts and ends with ```
        lang = element.lang or ''
        lang = lang.lower()
        if lang not in (
            "abap", "arduino",
            "bash", "basic",
            "c", "clojure", "coffeescript", "c++", "c#", "css",
            "dart", "diff", "docker",
            "elixir", "elm", "erlang",
            "flow", "fortran", "f#",
            "gherkin", "glsl", "go", "graphql", "groovy",
            "haskell", "html",
            "java", "javascript", "json", "julia",
            "kotlin",
            "latex", "less", "lisp", "livescript", "lua",
            "makefile", "markdown", "markup", "matlab", "mermaid",
            "nix",
            "objective-c", "ocaml",
            "pascal", "perl", "php", "plain text", "powershell", "prolog", "protobuf", "python",
            "r", "reason", "ruby", "rust",
            "sass", "scala", "scheme", "scss", "shell", "solidity", "sql", "swift",
            "typescript",
            "vb.net", "verilog", "vhdl", "visual basic",
            "webassembly",
            "xml",
            "yaml",
            "java/c/c++/c#",
        ):
            lang = {'js': 'javascript'}.get(lang) or ''
        lang = lang or 'plain text'  # set default
        return self._render_as('code', element, language=lang)

    def render_code_block(self, element):
        # https://developers.notion.com/reference/block#code-blocks
        return self.render_fenced_code(element)

    def render_html_block(self, element):
        # https://developers.notion.com/reference/block#code-blocks
        return self._render_as('code', element, language='html')

    def render_thematic_break(self, element):
        # https://developers.notion.com/reference/block#divider-blocks
        return self._render_as('divider', None)

    def render_heading(self, element):
        # https://developers.notion.com/reference/block#heading-one-blocks
        if element.level <= 3:
            return self._render_as(f'heading_{element.level}', element)
        raise NotImplementedError

    def render_setext_heading(self, element):
        return self.render_heading(element)

    def render_blank_line(self, element):
        return self._render_as("paragraph", '')

    def render_link_ref_def(self, element):
        # TODO: quick bypass
        return self._render_as("paragraph", '')
        # raise NotImplementedError

    def render_emphasis(self, element):
        return self._text(element, italic=True)

    def render_strong_emphasis(self, element):
        return self._text(element, bold=True)

    def render_inline_html(self, element):
        return self._text(element, code=True)

    def render_link(self, element):
        # https://developers.notion.com/reference/rich-text#text-objects
        return self._text(element, url=element.dest)

    def render_auto_link(self, element):
        # https://developers.notion.com/reference/rich-text#link-objects
        return self._text(element.dest, url=element.dest)

    def render_image(self, element):
        # https://developers.notion.com/reference/block#image-blocks
        return {
            'type': 'image',
            'image': {
                'type': 'external',
                'external': element.dest
            }
        }
        # 'value': self.render_children(element),
        # 'title': element.title,

    def render_literal(self, element):
        return self._text(element)

    def render_raw_text(self, element):
        return self._text(element)

    def render_line_break(self, element):
        return self._text('\n')
        # return {
        #     'type': 'br',
        #     'soft': element.soft,
        # }

    def render_code_span(self, element):
        # code span is inline code
        return self._text(element, code=True)


class MarkoNotionExt:
    # see https://github.com/frostming/marko/blob/master/marko/ext/gfm/__init__.py#L79
    elements = [Strikethrough]


_md = marko.Markdown(marko.Parser, MarkoNotionRenderer)
_md.use(MarkoNotionExt)

def md(text):
    result = _md(text)
    # flattn first level list, for list items inside a list block
    def iteritems():
        for idx, item in enumerate(result):
            if isinstance(item, list):
                yield from item
                continue
            if not isinstance(item, dict):
                print(f'invalid format {idx}/{len(result)}', json.dumps(result, indent=4))
                raise ValueError('format')
            yield item
    return list(iteritems())

def md_line(text):
    """for title"""
    return pydash.get(md(text), '0.paragraph.rich_text') or text



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="notion-params",
    version="0.0.5",
    author="Wenbo Zhao",
    author_email="zhaowb@gmail.com",
    description="Helper to build Notion API params, parse markdown text into Notion API blocks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zhaowb/notion-params.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'marko==1.2.0',
        'pydash',
        'requests',
        'backoff',
    ],
    py_modules=['notion_params']
)
